/*
   $Id: find.c,v 1.5 2002/12/28 09:18:58 bhepple Exp $

   part of gjots - a gnome/X11 jotter

   Copyright (C) 2002 Robert Hepple

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA 02111-1307, USA.

*/

/*
 * doFindNext	- if lastHit then start after it else current pt, current pt = sop
 * doReplace	- if lastHit then start there else current pt, current pt = eop
 * doReplaceAll	- if lastHit then start there else current pt, current pt = eop
 */

#include <ctype.h>
#include <gnome.h>
#include <regex.h>

#include "callbacks.h"
#include "interface.h"
#include "support.h"
#include "main.h"
#include "tree.h"
#include "file.h"
#include "find.h"

struct hit {
	GtkWidget	*item;
	regmatch_t 	match[10];
} lastHit;

static GtkWidget 		*findDialog = 0;
static GtkEntry 		*findString, *replaceString;
static GtkLabel			*findFeedback;
static GtkCheckButton	*caseCheckbutton, *backwardsCheckbutton,
	*regexCheckbutton, *globalCheckbutton;

static int 				findValue;
static regex_t 			regex;

/* like strchr() but ignore case */
static char *
strcasechr(char *haystack, char needle) {
	char ch = toupper(needle);
	for (; toupper(*haystack) != ch; ++haystack)
		if (*haystack == 0)
			return NULL;
	return haystack;
}

/* like strstr() but ignore case */
static char *
strcasestr(char *haystack, char *needle) {
	if (*needle == 0)
		return(haystack);

	for (; (haystack = strcasechr(haystack, *needle)) != NULL; ++haystack) {
		char *sc1 = haystack, *sc2 = needle;
		while (1) {
			if (*++sc2 == 0)
				return(haystack);
			else if (toupper(*++sc1) != toupper(*sc2))
				break;
		}
	}
	return NULL;
}

/*
 * re_subs:
 *	substitute the matched portions of the newPattern in dest.
 *
 *	\&		substitute the entire matched pattern.
 *
 * \<digit> substitute a subpattern, with the given tag number. Tags are
 *		    numbered from 1 to 9. If the particular tagged subpattern
 *		    does not exist, null is substituted. * Tag 0 == \&
 *
 *  \U \L   switch all following text to Upper/Lower case
 *  \E      stop \U \L
 *  \Q      quote till \E
 *  \u \l   next char is upper/lower case
 *  \n      <CR>
 *  \\      \
 */

static int
re_subs(char *orig, char *newPattern, char *dest, regmatch_t *match)
{
	char 	c;
	int  	matchNum;
	char 	*patternStart, *patternEnd;
	int 	shiftToUpper = 0, shiftToLower = 0, quoteMode = 0, once = 0;

	if (!newPattern || !dest)
		return(0);

	*dest = 0;
	if (!*newPattern || match[0].rm_so == -1 || match[0].rm_eo == -1)
		return 0;

	while ((c = *newPattern++)) {
		switch(c) {

		case '\\':
			c = *newPattern++;
			if (!c) {
				*dest = 0;
				return 1;
			}
			if (c == 'E') {
				shiftToUpper = shiftToLower = quoteMode = once = 0;
				continue;
			}
			if (quoteMode) {
				*dest++ = '\\';
				*dest++ = c;
				continue;
			}
			if (c >= '0' && c <= '9') {
				matchNum = c - '0';
				break;
			} else if (c == '&') {
				matchNum = 0;
				break;
			} else {
				switch (c) {
				case '\\': *dest++ = c;      once = 0;				continue;
				case 'n':  *dest++ = '\n';   once = 0;				continue;
				case 'l':  shiftToLower = 1; once = 1;				continue;
				case 'u':  shiftToUpper = 1; once = 1;				continue;
				case 'U':  shiftToUpper = 1; once = 0;				continue;
				case 'L':  shiftToLower = 1; once = 0;				continue;
				default:
					*dest++ = (shiftToUpper? toupper(c):
							  (shiftToLower? tolower(c):
							   c));
					if (once)
						shiftToLower = shiftToUpper = once = 0;
					continue;
				}
				continue;
			}
		default: /* not a \ */
			*dest++ = (shiftToUpper? toupper(c):
					  (shiftToLower? tolower(c):
					   c));
			if (once)
				shiftToLower = shiftToUpper = once = 0;
			continue;
		}

		/* copy pattern n: */
		if ((patternStart = orig + match[matchNum].rm_so) && 
			(patternEnd =   orig + match[matchNum].rm_eo)) {
			while (*patternStart && patternStart < patternEnd) {
				*dest++ = (shiftToUpper? toupper(*patternStart++):
						    (shiftToLower? tolower(*patternStart++):
							       *patternStart++));
				if (once)
					shiftToLower = shiftToUpper = once = 0;
			}
			if (patternStart < patternEnd) {
				*dest = 0;
				return(0);
			}
		}
	}
	*dest = (char) 0;
	return(1);
}

/********* FIND DIALOG *********/

static void
saveFindSettings() {
	if (findDialog) {
		safe_strncpy(prefs.findString, gtk_entry_get_text(findString), MAXSTRING);
		safe_strncpy(prefs.replaceString, gtk_entry_get_text(replaceString), MAXSTRING);
		prefs.matchCase = gtk_toggle_button_get_active(
			GTK_TOGGLE_BUTTON(caseCheckbutton));
		prefs.backwardsSearch = gtk_toggle_button_get_active(
			GTK_TOGGLE_BUTTON(backwardsCheckbutton));
		prefs.findStringIsRegex = gtk_toggle_button_get_active(
			GTK_TOGGLE_BUTTON(regexCheckbutton));
		prefs.findIsGlobal = gtk_toggle_button_get_active(
			GTK_TOGGLE_BUTTON(globalCheckbutton));
	}
}

static char *
getRegexErrorString(int err) {
	switch (err) {
	case REG_BADRPT:
		return "Invalid use of repetition operators such as using `*' as the first character.";
		break;

	case REG_BADBR:
		return "Invalid use of back reference operator.";
		break;

	case REG_EBRACE:
		return "Un-matched brace interval operators.";
		break;

	case REG_EBRACK:
		return "Un-matched bracket list operators.";
		break;

	case REG_ERANGE:
		return "Invalid use of the range operator, eg. the ending point of the range occurs prior to the starting point.";
		break;

	case REG_ECTYPE:
		return "Unknown character class name.";
		break;

	case REG_ECOLLATE:
		return "Invalid collating element.";
		break;

	case REG_EPAREN:
		return "Un-matched parenthesis group operators.";
		break;

	case REG_ESUBREG:
		return "Invalid back reference to a subexpression.";
		break;

	case REG_EEND:
		return "Non specific error.";
		break;

	case REG_EESCAPE:
		return "Trailing backslash.";
		break;

	case REG_BADPAT:
		return "Invalid use of pattern operators such as group or list.";
		break;

	case REG_ESIZE:
		return "Compiled regular expression requires a pattern buffer larger than 64Kb.";
		break;

	case REG_ESPACE:
		return "The regex routines ran out of memory.";
		break;
	}
	return "";
}

/* puts result into global variable lastHit */
static int
findNext(GtkWidget *startItem, int startPos) {

	struct payload_s 	*payload;
	char 				*found = 0;
	GtkWidget			*item;
	regmatch_t			regmatch[10];
	int					i;

	item = startItem;

	lastHit.item = NULL;
	for (i = 0; i < 10; i++) {
		lastHit.match[i].rm_so = lastHit.match[i].rm_eo = -1;
	}

	do {
		if (!item) 
			return 0;
		payload = (struct payload_s *)gtk_object_get_user_data(GTK_OBJECT(item));
		if (!payload || !payload->text) return 0;

		if (strlen(payload->text) >= startPos) {
			if (prefs.findStringIsRegex) {
				if (regexec(&regex, payload->text + startPos, 10, regmatch, 0) == 0) {
					found = payload->text + startPos + regmatch[0].rm_so;
					lastHit.item = item;
					for (i = 0; i < 10 && regmatch[i].rm_so >= 0; i++) {
						lastHit.match[i].rm_so = startPos + regmatch[i].rm_so;
						lastHit.match[i].rm_eo = startPos + regmatch[i].rm_eo;
					}
				}
			} else { /* seek plain text, not regex */
				found = prefs.matchCase?
					strstr(payload->text + startPos, prefs.findString):
					strcasestr(payload->text + startPos, prefs.findString);
				if (found) {
					lastHit.item = item;
					lastHit.match[0].rm_so = found - payload->text;
					lastHit.match[0].rm_eo = found - payload->text + strlen(prefs.findString);
				}
			}
		}

		/* try the next item: */
		if (!found) {
			GtkWidget *thisItem;

			if (!prefs.findIsGlobal) {
				return 0;
			}

			thisItem = getNextItem(item, ABSOLUTE); /* returns last item if at last item! */
			if (thisItem == NULL || thisItem == item)
				item = NULL;
			else
				item = thisItem;
			startPos = 0;
		}
	} while (!found && item != NULL);
	return ((int)found);
}

void
doFindNext(int update_now) {
	GtkWidget	*startItem;
	int 		startPos;

	syncPayload();
	saveFindSettings();
	if (findDialog)
		gtk_label_set_text(findFeedback, "");

	if (lastHit.item) {
		startItem = lastHit.item;
		startPos = lastHit.match[0].rm_eo;
	} else {
		startItem = (GtkWidget *) 
			gtk_object_get_user_data(GTK_OBJECT(textDisplay));
		startPos = 
			gtk_editable_get_position(GTK_EDITABLE(textDisplay));
	}

	if (prefs.findStringIsRegex) {
		int error;

		error = regcomp(&regex, 
						prefs.findString, 
						REG_EXTENDED | REG_NEWLINE |
						(prefs.matchCase? 0: REG_ICASE));
		if (error) {
			if (findDialog)
				gtk_label_set_text(findFeedback, getRegexErrorString(error));
			msg(getRegexErrorString(error));
			regfree(&regex);
			lastHit.item = NULL;
			return;
		}
	}

	if (findNext(startItem, startPos)) {
		if (lastHit.item != startItem) {
			unselectAll();
			// gtk_widget_grab_focus(newItem);
			// gtk_widget_grab_default(newItem);
			gtk_tree_select_child((GtkTree *)lastHit.item->parent, lastHit.item);
			checkOrigin(lastHit.item);
		}
		gtk_editable_set_position(GTK_EDITABLE(textDisplay), 
								  lastHit.match[0].rm_so);
		gtk_editable_select_region(GTK_EDITABLE(textDisplay), 
								   lastHit.match[0].rm_so, 
								   lastHit.match[0].rm_eo);
	} else {
		if (findDialog)
			gtk_label_set_text(findFeedback, "Not found");
		msg("Not found");
	}

	if (prefs.findStringIsRegex)
		regfree(&regex);
}

void
unPopFind(int isOK) 
{
	if (isOK) {
		findValue = OK;
		saveFindSettings();
	} else
		findValue = CANCEL;
	gtk_widget_destroy(findDialog);
}

void
zapFind()
{
	findDialog = 0;
}

static int
replaceNext(GtkWidget *startItem, int startPos) {
	char 				*tmpText, *dest, *src, *title_end;
	struct payload_s 	*payload;
	int					newSize, new_title = 0;

	if (readonly(READONLY_ENQ))
		return 0;

	findNext(startItem, startPos);

	/* by now we should have a hit */
	if (lastHit.item == NULL || 
		lastHit.match[0].rm_so == -1 ||
		lastHit.match[0].rm_eo == -1) {
		if (prefs.findStringIsRegex)
			regfree(&regex);

		if (findDialog)
			gtk_label_set_text(findFeedback, "Not found");
		msg("Not found");

		return 0;
	}

	payload = (struct payload_s *)gtk_object_get_user_data(GTK_OBJECT(lastHit.item));
	if (!payload || !payload->text) {
		if (prefs.findStringIsRegex)
			regfree(&regex);
		return 0;
	}

	title_end = strchr(payload->text, '\n');
	if (title_end == NULL)
		title_end = payload->text + strlen(payload->text);

	new_title = (lastHit.match[0].rm_so < (title_end - payload->text));

	/* Now create a new payload->text with the new values: This is a
	 * guess at the size - could be wrong, hmmmm, need to make this
	 * safer: */
	newSize = strlen(payload->text) * 2;
	if (newSize < 1024)
		newSize = 1024;

	tmpText = alloca(newSize);
	src = payload->text;
	dest = tmpText;

	/* text is llllPPPPrrrr: copy leftmost part of text: */
	if (lastHit.match[0].rm_so)
		strncpy(dest, src, lastHit.match[0].rm_so);
	dest += lastHit.match[0].rm_so;

	/* copy pattern match: */
	if (prefs.findStringIsRegex)
		re_subs(payload->text, prefs.replaceString, dest, lastHit.match);
	else
		strcpy(dest, prefs.replaceString);
	dest += strlen(dest);

	/* copy right part of text: */
	src = payload->text + lastHit.match[0].rm_eo;
	strcpy(dest, src);

	/* update cursor value to new end of hit: */
	lastHit.match[0].rm_eo = dest - tmpText;

	free(payload->text);
	payload->text = malloc(strlen(tmpText) + 1);
	strcpy(payload->text, tmpText);

	if (new_title)
		checkTitle(lastHit.item);
	dirty_data = 1;
	return 1;
}

void
doReplace()
{
	GtkWidget 			*startItem;
	int					startPos;
	struct payload_s	*payload;

	syncPayload();
	saveFindSettings();
	if (findDialog)
		gtk_label_set_text(findFeedback, "");

	startItem = (GtkWidget *) 
		gtk_object_get_user_data(GTK_OBJECT(textDisplay));

	if (lastHit.item) {
		startItem = lastHit.item;
		startPos = lastHit.match[0].rm_so;
	} else {
		startItem = (GtkWidget *) 
			gtk_object_get_user_data(GTK_OBJECT(textDisplay));
		startPos = 
			gtk_editable_get_position(GTK_EDITABLE(textDisplay));
	}

	if (prefs.findStringIsRegex) {
		int error;

		error = regcomp(&regex, 
						prefs.findString, 
						REG_EXTENDED | REG_NEWLINE |
						(prefs.matchCase? 0: REG_ICASE));
		if (error) {
			if (findDialog)
				gtk_label_set_text(findFeedback, getRegexErrorString(error));
			msg(getRegexErrorString(error));
			regfree(&regex);
			lastHit.item = NULL;
			return;
		}
	}

	replaceNext(startItem, startPos);

	if (lastHit.item) {
		/* if we're now on a different item then display it: */
		if (lastHit.item != startItem) {
			unselectAll();
			// gtk_widget_grab_focus(savedHit.item);
			// gtk_widget_grab_default(savedHit.item);
			gtk_tree_select_child((GtkTree *)lastHit.item->parent, lastHit.item);
			checkOrigin(lastHit.item);
		}

		gtk_text_freeze(textDisplay);
		clearTextDisplay();

		payload = (struct payload_s *)gtk_object_get_user_data(GTK_OBJECT(lastHit.item));
		gtk_text_insert(textDisplay, textDisplayFont, NULL, NULL, payload->text, -1);

		gtk_editable_set_position(GTK_EDITABLE(textDisplay), lastHit.match[0].rm_eo);
		gtk_editable_select_region(GTK_EDITABLE(textDisplay), 
								   lastHit.match[0].rm_so, 
								   lastHit.match[0].rm_eo);
		gtk_text_thaw(textDisplay);
	}
	lastHit.item = NULL;

	if (prefs.findStringIsRegex)
		regfree(&regex);

	return;
}

void
doReplaceAll()
{
	GtkWidget 			*startItem, *thisItem;
	int					startPos, thisPos;
	struct hit 			savedHit;
	int					hit_count = 0;
	struct payload_s 	*payload;

	syncPayload();
	saveFindSettings();
	if (findDialog)
		gtk_label_set_text(findFeedback, "");

	if (lastHit.item) {
		startItem = lastHit.item;
		startPos = lastHit.match[0].rm_so;
	} else {
		startItem = (GtkWidget *) 
			gtk_object_get_user_data(GTK_OBJECT(textDisplay));
		startPos = 
			gtk_editable_get_position(GTK_EDITABLE(textDisplay));
	}

	savedHit.item = NULL;

	if (prefs.findStringIsRegex) {
		int error;

		error = regcomp(&regex, 
						prefs.findString, 
						REG_EXTENDED | REG_NEWLINE |
						(prefs.matchCase? 0: REG_ICASE));
		if (error) {
			if (findDialog)
				gtk_label_set_text(findFeedback, getRegexErrorString(error));
			msg(getRegexErrorString(error));
			regfree(&regex);
			lastHit.item = NULL;
			return;
		}
	}

	thisItem = startItem;
	thisPos = startPos;
	while (replaceNext(thisItem, thisPos)) {
		memcpy((char *)&savedHit, (char *)&lastHit, sizeof(struct hit));
		hit_count++;
		thisItem = lastHit.item;
		thisPos = lastHit.match[0].rm_eo;
	}

	if (hit_count && savedHit.item) {
		/* if we're now on a different item then display it: */
		if (savedHit.item != startItem) {
			unselectAll();
			// gtk_widget_grab_focus(savedHit.item);
			// gtk_widget_grab_default(savedHit.item);
			gtk_tree_select_child((GtkTree *)savedHit.item->parent, savedHit.item);
			checkOrigin(savedHit.item);
		}

		/* set the textDisplay */
		gtk_text_freeze(textDisplay);
		clearTextDisplay();
		payload = (struct payload_s *)
			gtk_object_get_user_data(GTK_OBJECT(savedHit.item));
		gtk_text_insert(textDisplay, textDisplayFont, NULL, NULL, payload->text, -1);
		gtk_editable_set_position(GTK_EDITABLE(textDisplay), savedHit.match[0].rm_eo);
		gtk_editable_select_region(GTK_EDITABLE(textDisplay), 
								   savedHit.match[0].rm_so, 
								   savedHit.match[0].rm_eo);
		gtk_text_thaw(textDisplay);
	}

	{
		char buf[MAXBUF];

		sprintf(buf, "%d replacements made", hit_count);
		if (findDialog)
			gtk_label_set_text(findFeedback, buf);
		msg(buf);
	}

	if (prefs.findStringIsRegex)
		regfree(&regex);
}

void
clearFind()
{
	lastHit.item = NULL;
}

void
popupFind() {

	if (findDialog) {
		unPopFind(1);
		return;
	}

	findDialog = create_findDialog();

	findString =    (GtkEntry *)lookup_widget(findDialog, "findEntry");
	replaceString = (GtkEntry *)lookup_widget(findDialog, "replaceEntry");
	caseCheckbutton = (GtkCheckButton *)lookup_widget(findDialog, "caseCheckbutton");
	backwardsCheckbutton = (GtkCheckButton *)lookup_widget(findDialog, "backwardsCheckbutton");
	regexCheckbutton = (GtkCheckButton *)lookup_widget(findDialog, "regexCheckbutton");
	globalCheckbutton = (GtkCheckButton *)lookup_widget(findDialog, "globalCheckbutton");
	findFeedback = (GtkLabel *)lookup_widget(findDialog, "findFeedback");

	if (!findString ||
		!replaceString ||
		!caseCheckbutton ||
		!backwardsCheckbutton ||
		!regexCheckbutton ||
		!globalCheckbutton ||
		!findFeedback)
		return;


	gtk_entry_set_text(findString, prefs.findString);
	gtk_entry_set_text(replaceString, prefs.replaceString);
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(caseCheckbutton),
								 prefs.matchCase);
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(backwardsCheckbutton),
								 prefs.backwardsSearch);
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(regexCheckbutton),
								 prefs.findStringIsRegex);
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(globalCheckbutton),
								 prefs.findIsGlobal);
	gtk_label_set_text(findFeedback, "");

	gtk_widget_show(findDialog);
}

void
doFindClear()
{
	gtk_entry_set_text(findString, "");
	gtk_entry_set_text(replaceString, "");
}
